#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>

#define COUNT 5000000  

int n = 0;

struct lock_st {
	int flag;
};

typedef struct lock_st lock_t;

lock_t lock;

// Cette fonction initialise le verrou passé en argument
void init_lock(lock_t *lock) {
	lock->flag = 0;
}

// Cette fonction verrouille (acquiert) le verrou passé en argument
void acquire_lock(lock_t *lock) {
	while (__sync_val_compare_and_swap(&lock->flag, 0, 1));
	
}

// Cette fonction déverrouille (relâche) le verrou passé en argument
void release_lock(lock_t *lock) {
	__sync_val_compare_and_swap(&lock->flag, 1, 0);
}

void *compute(void *arg) {
	int i;
	for (i = 0; i < COUNT; i++) {
		// critical section
		acquire_lock(&lock);
		n++;
		release_lock(&lock);
		// end of critical section
	}

	return NULL;
}

void error(char *msg) {
	perror(msg);
	exit(EXIT_FAILURE);
}

int main(int argc, char **argv) {
	pthread_t thread1, thread2;

	init_lock(&lock);

	if (pthread_create(&thread1, NULL, compute, NULL) != 0)
		error("pthread_create");
	if (pthread_create(&thread2, NULL, compute, NULL) != 0)
		error("pthread_create");

	if (pthread_join(thread1, NULL) != 0)
		error("pthread_join");
	if (pthread_join(thread2, NULL) != 0)
		error("pthread_join");

	int errors = 2*COUNT-n;
	if (errors)
		printf("%d ERRORS out of %d concurrent iterations!\n", errors, COUNT);
	else
		printf("OK: 0 errors in %d concurrent iterations\n", COUNT);

	return EXIT_SUCCESS;
}
